@@ -30,25 +30,30 @@ SEXP _locate(SEXP sym, SEXP env, SEXP function) {
 
   while (env != R_EmptyEnv) {
     assert_type(env, ENVSXP);
+    LOG("looking in env %p for %s", env, CHAR(PRINTNAME(sym)));
     if (fn) {
       SEXP x = findVarInFrame3(env, sym, TRUE);
-      if (TYPEOF(x) == PROMSXP) {
+      LOG("got a %s", type2char(TYPEOF(x)));
+      while (TYPEOF(x) == PROMSXP) {
         if (PRVALUE(x) == R_UnboundValue) {
           /* Per R rules, we must force. As forcing isn't exposed in
              Rinternals, I'll do it by calling "force"... or forceAndCall and
              then calling force.... */
+          LOG("forcing it");
           SEXP force = findVarInFrame3(R_BaseNamespace, install("force"), TRUE);
           SEXP callForce = PROTECT(list2(force, sym));
           R_forceAndCall(callForce, 1, env);
           UNPROTECT(1);
-          x = PRVALUE(x);
         }
+        x = PRVALUE(x);
+        LOG("containing a %s", type2char(TYPEOF(x)));
       }
 
       switch(TYPEOF(x)) {
       case CLOSXP: 
       case SPECIALSXP:
       case BUILTINSXP:
+        LOG("found a %s", type2char(TYPEOF(x)));
         return env;
       default:
         break;
@@ -117,6 +122,12 @@ int unwrappable(SEXP prom) {
   return unwrappable;
 }
 
+int is_forced_promise(SEXP prom) {
+  return !( PRVALUE(prom) == R_UnboundValue
+            || PRVALUE(prom) == R_MissingArg
+            || PRENV(prom) == NULL);
+}
+
 SEXP unwrap_step(SEXP prom) {
   // attmpt to unwrap, returning another promise if one is found.
   // If no other promise is found return R_UnboundValue.
@@ -175,6 +186,44 @@ SEXP unwrap_promise(SEXP prom, int recursive) {
   return prom;
 }
 
+/* Attempt to tease out a value from a promise without forcing.
+   If unsuccessful return R_UnboundValue. */
+SEXP peek_promise(SEXP prom) {
+  LOG("got a %s", type2char(TYPEOF(prom)));
+  
+  prom = unwrap_promise(prom, TRUE);
+  assert_type(prom, PROMSXP);
+  while(TYPEOF(PREXPR(prom)) == PROMSXP) {
+    prom = PREXPR(prom);
+  }
+  LOG("unwrapped to a %s", type2char(TYPEOF(prom)));
+  if (is_forced_promise(prom)) {
+    LOG("it's forced - a %s", type2char(TYPEOF(PRVALUE(prom))));
+    return PRVALUE(prom);
+  }
+  SEXP name = PREXPR(prom);
+  SEXP env = PRENV(prom);
+  switch(TYPEOF(name)) {
+  case SYMSXP: {
+    LOG("it's a %s, `%s`, quoted in environment %p",
+        type2char(TYPEOF(name)), CHAR(PRINTNAME(name)), PRENV(prom));
+    SEXP binding;
+    binding = x_findVar(name, env);
+    // we've already unwrapped so this shouldn't be another promise
+    LOG("...bound to a  %s", type2char(TYPEOF(binding)));
+    assert(TYPEOF(binding) != PROMSXP);
+    return binding; // possibly R_UnboundValue
+  }
+  case LANGSXP:
+    LOG("it's a call");
+    return R_UnboundValue;
+  default:
+    LOG("it's a value");
+    // this must be a value
+    return name;
+  }
+}
+
 SEXP _unwrap_quotation(SEXP q, SEXP recursive) {
   SEXP pr = PROTECT(_quotation_to_promsxp(q));
   pr = PROTECT(unwrap_promise(pr, asLogical(recursive)));
@@ -503,13 +552,7 @@ SEXP arg_check(SEXP envir, SEXP name, TEST_ENUM query, int warn) {
     case IS_PROMISE: 
       return ScalarLogical(TRUE);
     case IS_FORCED:
-      if ( PRVALUE(binding) == R_UnboundValue
-           || PRVALUE(binding) == R_MissingArg
-           || PRENV(binding) == NULL) {
-        return ScalarLogical(FALSE);
-      } else {
-        return ScalarLogical(TRUE);
-      }
+      return is_forced_promise(binding) ? ScalarLogical(TRUE) : ScalarLogical(FALSE);
     }
   case SYMSXP:
     switch(query) {
